@@ -120,6 +120,7 @@ struct microcode_bundle {
 	const void *data;       /* binary file data */
 };
 static struct microcode_bundle *microcode_bundles = NULL;
+static struct microcode_bundle **microcode_bundles_tail = &microcode_bundles;
 static unsigned int next_bundle_id = 1;
 
 /* intel_uclist_entry signature flag constants */
@@ -140,7 +141,7 @@ enum {
 struct intel_uclist_entry {
 	struct intel_uclist_entry *next;
 	uint32_t cpuid;		/* sorting key */
-	uint32_t pf_mask;
+	uint32_t pfm;		/* microcode platform ID (MSR 0x17) mask */
 	int32_t  uc_rev;	/* duplicated from header */
 	uint32_t flags;		/* INTEL_UCLE_ flags */
 	unsigned int id;	/* microcode id within group */
@@ -152,6 +153,7 @@ static struct intel_uclist_entry *all_microcodes = NULL;
 static struct intel_uclist_entry *microcodes = NULL;
 
 /* UI helpers */
+#define UCODE_ID_FMT_UU "%02u/%03u"
 #define UCODE_ID_MAX_LEN 22
 struct microcode_iterator_data {
 	const struct microcode_bundle *current_bundle;
@@ -167,7 +169,7 @@ static struct microcode_iterator_data microcode_iterator_data;
 struct microcode_filter_entry {
 	struct microcode_filter_entry *next;
 	uint32_t cpuid;		/* exact match */
-	uint32_t pf_mask;	/* common bits set match */
+	uint32_t pfm;		/* common bits set match */
 	int invert;
 };
 static struct microcode_filter_entry *uc_filter_list = NULL;
@@ -206,6 +208,9 @@ static unsigned int extsig_tables_in_use = 0;
 #define print_err(format, arg...) \
 	do { fflush(stdout); fprintf(stderr, "%s: " format "\n", progname, ## arg); } while (0)
 
+#define print_warn(format, arg...) \
+	do { fflush(stdout); fprintf(stderr, "%s: warning: " format "\n", progname, ## arg); } while (0)
+
 static inline int is_dash(const char * const fn)
 {
 	return fn && *fn == '-' && !*(fn+1);
@@ -260,7 +265,7 @@ static void str_append_ucode_id(unsigned int id,
 				unsigned int gid,
 				char *str, unsigned int len)
 {
-	snprintf(str, len, "%02u/%03u", gid, id);
+	snprintf(str, len, UCODE_ID_FMT_UU, gid, id);
 }
 
 /* Signature linked list */
@@ -296,8 +301,9 @@ static void xx_uclist_remove_shadowed(const unsigned int gid,
 	while (*pnext && (*pnext)->cpuid == cpuid) {
 		struct intel_uclist_entry * const e = *pnext;
 
-		if (((pf_mask & e->pf_mask) == e->pf_mask) &&
-		    (uc_rev >= e->uc_rev || (downgrade && gid > e->gid))) {
+		if (((pf_mask & e->pfm) == e->pfm) &&
+		    ((!downgrade && uc_rev >= e->uc_rev) ||
+		     (downgrade && gid > e->gid))) {
 			/* entry is a subset of the one we just added */
 			/* delete entry */
 			*pnext = e->next;
@@ -363,7 +369,7 @@ static int uclist_add_signature(const unsigned int id,
 
 	/* look for duplicates, stop on the first */
 	while (likely(*pnext) && (*pnext)->cpuid == cpuid) {
-		if ((*pnext)->pf_mask == pf_mask && (*pnext)->uc_rev == uc_rev) {
+		if ((*pnext)->pfm == pf_mask && (*pnext)->uc_rev == uc_rev) {
 			res = EEXIST;
 			flags |= INTEL_UCLE_DUPSIG;
 			d = *pnext;
@@ -386,7 +392,7 @@ static int uclist_add_signature(const unsigned int id,
 	n->id = id;
 	n->gid = gid;
 	n->cpuid = cpuid;
-	n->pf_mask = pf_mask;
+	n->pfm = pf_mask;
 	n->uc_rev = uc_rev;
 	n->uc = uc;
 	n->uc_size = uc_size;
@@ -460,23 +466,23 @@ static int uclist_merge_signature(const unsigned int id,
 	while (*pnext && (*pnext)->cpuid == cpuid) {
 		struct intel_uclist_entry * const e = *pnext;
 
-		if (pf_mask && (e->pf_mask & pf_mask) == 0) {
+		if (pf_mask && (e->pfm & pf_mask) == 0) {
 			/* disjoint sets, continue search */
 			pnext = &(e->next);
 		} else if (!downgrade || gid == e->gid) { /* do not downgrade */
-			if ((e->pf_mask & pf_mask) == pf_mask) {
+			if ((e->pfm & pf_mask) == pf_mask) {
 				/* new set same or contained in the old set */
 				if (e->uc_rev >= uc_rev) {
 					/* old set as good, or newer */
 					return EEXIST;
-				} else if (pf_mask == e->pf_mask) {
+				} else if (pf_mask == e->pfm) {
 					/* the sets are identical, and the new
 					 * one is newer microcode: replace */
 					e->id = id;
 					e->gid = gid;
 					e->flags = flags;
 					/* e->cpuid = cpuid; */
-					e->pf_mask = pf_mask;
+					e->pfm = pf_mask;
 					e->uc_rev = uc_rev;
 					e->uc = uc;
 					e->uc_size = uc_size;
@@ -486,15 +492,15 @@ static int uclist_merge_signature(const unsigned int id,
 						     downgrade, &(e->next));
 					return 0;
 				}
-			} else if ((pf_mask & e->pf_mask) == e->pf_mask) {
-				/* new entry a superset of the old */
+			} else if ((pf_mask & e->pfm) == e->pfm) {
+				/* new set a superset of the old */
 				if (uc_rev >= e->uc_rev) {
 					/* we can replace the old entry */
 					e->id = id;
 					e->gid = gid;
 					e->flags = flags;
 					/* e->cpuid = cpuid; */
-					e->pf_mask = pf_mask;
+					e->pfm = pf_mask;
 					e->uc_rev = uc_rev;
 					e->uc = uc;
 					e->uc_size = uc_size;
@@ -507,14 +513,14 @@ static int uclist_merge_signature(const unsigned int id,
 			}
 			pnext = &(e->next);
 		} else if (gid > e->gid) { /* downgrade */
-			if (e->pf_mask == pf_mask) {
+			if (e->pfm == pf_mask) {
 				int rc = (e->uc_rev == uc_rev) ? EEXIST : 0;
 				/* same set, replace */
 				e->id = id;
 				e->gid = gid;
 				e->flags = flags;
 				/* e->cpuid = cpuid; */
-				e->pf_mask = pf_mask;
+				e->pfm = pf_mask;
 				e->uc_rev = uc_rev;
 				e->uc = uc;
 				e->uc_size = uc_size;
@@ -523,7 +529,7 @@ static int uclist_merge_signature(const unsigned int id,
 					xx_uclist_remove_shadowed(gid, cpuid, pf_mask, uc_rev,
 								  downgrade, &(e->next));
 				return rc;
-			} else if ((e->pf_mask & pf_mask) == pf_mask && e->uc_rev == uc_rev) {
+			} else if ((e->pfm & pf_mask) == pf_mask && e->uc_rev == uc_rev) {
 				/* new set contained in the old set, and same rev */
 				return EEXIST;
 			} else {
@@ -531,7 +537,7 @@ static int uclist_merge_signature(const unsigned int id,
 				break;
 			}
 		} else { /* downgrade mode but older group and not disjoint */
-			if ((e->pf_mask & pf_mask) == pf_mask)
+			if ((e->pfm & pf_mask) == pf_mask)
 				return EEXIST; /* new set same or contained in the old set */
 			pnext = &(e->next);
 		}
@@ -546,7 +552,7 @@ static int uclist_merge_signature(const unsigned int id,
 	n->gid = gid;
 	n->flags = flags;
 	n->cpuid = cpuid;
-	n->pf_mask = pf_mask;
+	n->pfm = pf_mask;
 	n->uc_rev = uc_rev;
 	n->uc = uc;
 	n->uc_size = uc_size;
@@ -579,12 +585,11 @@ static int add_intel_microcode_bundle(const char * const mcb_filename,
 				      const void * const mcb,
 				      const size_t mcb_size)
 {
-	struct microcode_bundle *b, **p;
-
 	assert(mcb && mcb_size >= 1024);
 	assert(mcb_filename);
+	assert(microcode_bundles_tail && !(*microcode_bundles_tail));
 
-	b = malloc(sizeof(struct microcode_bundle) + mcb_size);
+	struct microcode_bundle *b = malloc(sizeof(struct microcode_bundle));
 	if (!b)
 		return ENOMEM;
 
@@ -595,10 +600,8 @@ static int add_intel_microcode_bundle(const char * const mcb_filename,
 	b->data = mcb;
 
 	/* add to end of list */
-	p = &microcode_bundles;
-	while (*p)
-		p = &((*p)->next);
-	*p = b;
+	*microcode_bundles_tail = b;
+	microcode_bundles_tail = &(b->next);
 
 	return 0;
 }
@@ -618,6 +621,7 @@ static void free_intel_microcode_bundles(void)
 	}
 
 	microcode_bundles = NULL;
+	microcode_bundles_tail = &microcode_bundles;
 }
 
 static int load_intel_microcode_dat(FILE *fp, void **mcb, size_t *mcb_length,
@@ -1155,6 +1159,7 @@ static int load_intel_microcode(const char * path,
 			ftype = INTEL_UC_FT_BIN;
 
 		err = load_intel_microcode_file(fd, NULL, fn, ftype);
+		fd = -1; /* fd closed by load_intel_microcode_file() */
 	} while (dir && (!err || ignore_bad_ucode));
 
 	if (dir)
@@ -1192,12 +1197,10 @@ static void log_microcode_action(const char * const action,
 				const char * const devname,
 				const struct intel_uclist_entry * const uce)
 {
-	char id_str[UCODE_ID_MAX_LEN];
-
-	str_append_ucode_id(uce->id, uce->gid, id_str, sizeof(id_str));
-	print_msg_u("%s: %s microcode %s (sig 0x%08x, pf_mask 0x%02x, rev 0x%04x)",
-		    devname, action, id_str,
-		    uce->cpuid, uce->pf_mask, uce->uc_rev);
+	print_msg_u("%s: %s microcode " UCODE_ID_FMT_UU
+		    " (sig 0x%08x, pf_mask 0x%02x, rev 0x%04x)",
+		    devname, action, uce->gid, uce->id,
+		    uce->cpuid, uce->pfm, uce->uc_rev);
 }
 
 /**
@@ -1574,12 +1577,12 @@ static void xx_merge_filter(struct microcode_filter_entry * const f,
 			 const uint32_t pf_mask, const int invert)
 {
 	if (f->invert == invert) {
-		f->pf_mask |= pf_mask;
+		f->pfm |= pf_mask;
 	} else if (!(f->invert)) {
-		f->pf_mask &= ~pf_mask;
+		f->pfm &= ~pf_mask;
 	} else {
 		f->invert = 0;
-		f->pf_mask = pf_mask;
+		f->pfm = pf_mask;
 	}
 }
 
@@ -1599,7 +1602,7 @@ static int add_filter_to_list(uint32_t cpuid, uint32_t pf_mask, int invert,
 		pnext = &((*pnext)->next);
 
 	if (*pnext && (*pnext)->cpuid == cpuid) {
-		if (((pf_mask & (*pnext)->pf_mask) == pf_mask) &&
+		if (((pf_mask & (*pnext)->pfm) == pf_mask) &&
 		     ((*pnext)->invert == invert)) {
 			/* found equivalent, report it */
 			return EEXIST;
@@ -1614,7 +1617,7 @@ static int add_filter_to_list(uint32_t cpuid, uint32_t pf_mask, int invert,
 	if (!n)
 		return ENOMEM;
 	n->cpuid = cpuid;
-	n->pf_mask = pf_mask;
+	n->pfm = pf_mask;
 	n->invert = invert;
 	n->next = *pnext;
 	*pnext = n;
@@ -1644,7 +1647,7 @@ static void add_filter_list_to_list(
 			pp = &((*pp)->next);
 
 		if (*pp && (*pp)->cpuid == e->cpuid) {
-			xx_merge_filter(*pp, e->pf_mask, e->invert);
+			xx_merge_filter(*pp, e->pfm, e->invert);
 		} else {
 			/* insert before */
 			e->next = *pp;
@@ -1663,8 +1666,8 @@ static int is_selected(uint32_t cpuid, uint32_t pf_mask,
 	while (f && f->cpuid < cpuid)
 		f = f->next;
 	while (f && f->cpuid == cpuid) {
-		if ((pf_mask & f->pf_mask) != 0 ||
-		    f->pf_mask == PFMASK_MATCH_ANY)
+		if ((pf_mask & f->pfm) != 0 ||
+		    f->pfm == PFMASK_MATCH_ANY)
 			return !(f->invert);
 		f = f->next;
 	}
@@ -1763,7 +1766,7 @@ static int xx_datefilter_loose_inplaceinsert(struct intel_uclist_entry **p,
 					     struct intel_uclist_entry **uclist)
 {
 	const uint32_t cpuid = (*p)->cpuid;
-	const uint32_t pfm   = (*p)->pf_mask;
+	const uint32_t pf_mask = (*p)->pfm;
 	struct intel_uclist_entry *e;
 	int rc;
 
@@ -1773,12 +1776,12 @@ static int xx_datefilter_loose_inplaceinsert(struct intel_uclist_entry **p,
 	*uclist = e; /* speed up next search */
 
 	while (e && e->cpuid == cpuid) {
-		if (e->flags & INTEL_UCLE_SELID && e->pf_mask & pfm) {
+		if (e->flags & INTEL_UCLE_SELID && e->pfm & pf_mask) {
 			e->flags &= ~INTEL_UCLE_SELID;
 			e->flags |= INTEL_UCLE_SELECT;
 			rc = xx_uclist_add_print_errors(
 			        uclist_merge_signature(e->id, e->gid,
-				        e->flags, e->cpuid, e->pf_mask,
+				        e->flags, e->cpuid, e->pfm,
 					e->uc_rev, e->uc, e->uc_size,
 					allow_downgrade, p));
 			if (rc)
@@ -1790,6 +1793,37 @@ static int xx_datefilter_loose_inplaceinsert(struct intel_uclist_entry **p,
 	return 0;
 }
 
+static int check_downgrade_shadowing(const struct intel_uclist_entry * uclist)
+{
+	int res = 0;
+
+	while (uclist) {
+		struct intel_uclist_entry *e = uclist->next;
+		uint32_t sig = uclist->cpuid;
+		uint32_t pfm = uclist->pfm;
+		int32_t  rev = uclist->uc_rev;
+		uint32_t gid = uclist->gid;
+
+		while(e && e->cpuid == sig) {
+			if ((pfm & e->pfm) != 0 &&
+			    ((gid > e->gid && rev < e->uc_rev) ||
+			     (gid < e->gid && rev > e->uc_rev))) {
+				print_warn("cannot downgrade microcode sig 0x%x, "
+					  "pf_mask 0x%02x (entries: " UCODE_ID_FMT_UU
+					  " rev 0x%0x, and " UCODE_ID_FMT_UU " rev 0x%0x)",
+					  sig, (pfm & e->pfm),
+					  gid, uclist->id, rev, e->gid, e->id, e->uc_rev);
+				res = EEXIST;
+			}
+			e = e->next;
+		}
+
+		uclist = uclist->next;
+	}
+
+	return res;
+}
+
 static int xx_process_ucode_signature_cb(void *userdata,
 				unsigned int const sig_count,
 				uint32_t const cpuid,
@@ -1853,7 +1887,7 @@ static int xx_process_ucode_signature_cb(void *userdata,
 	case EEXIST:
 		break;
 	case EBADF:
-		print_err("WARNING: Microcode %s has the same revision "
+		print_warn("microcode %s has the same revision "
 			  "and signature as a previously loaded microcode, "
 			  "but different contents", ctx->current_uc_id_str);
 		return 1;
@@ -1987,10 +2021,11 @@ static int do_process_microcodes(void)
 			if (list_sel_microcodes) {
 				struct intel_ucode_metadata m;
 
-				intel_ucode_getmetadata(uce->uc, &m);
+				if (unlikely(intel_ucode_getmetadata(uce->uc, &m) != INTEL_UCODE_NOERROR))
+					exit(EXIT_SWFAILURE); /* memory corruption */
 				printf("%03lu: sig 0x%08x, pf mask 0x%02x, "
 					"%04x-%02x-%02x, rev 0x%04x, size %u\n",
-					sigcount, uce->cpuid, uce->pf_mask,
+					sigcount, uce->cpuid, uce->pfm,
 					m.date_year, m.date_month, m.date_day,
 					m.revision, uce->uc_size);
 			}
@@ -2001,6 +2036,10 @@ static int do_process_microcodes(void)
 			  uccount, sigcount);
 	}
 
+	/* detect annoying downgrade mode shadows, and warn */
+	if (allow_downgrade)
+		check_downgrade_shadowing(microcodes);
+
 	return 0;
 }
 
@@ -2074,7 +2113,7 @@ static int do_write_named(const char * const dirname,
 	while (p && !rc) {
 		if (!(p->flags & INTEL_UCLE_DUPSIG)) {
 			snprintf(fn, sizeof(fn), "s%08X_m%08X_r%08X.fw",
-				     p->cpuid, p->pf_mask, p->uc_rev);
+				     p->cpuid, p->pfm, p->uc_rev);
 
 			memcpy(&e, p, sizeof(e));
 			e.next = NULL;
@@ -2148,7 +2187,7 @@ static int do_write_firmware(const char * const dirname)
 		cpuid = p->cpuid;
 		while (p && p->cpuid == cpuid) {
 			add_status = uclist_merge_signature(p->id, p->gid,
-						p->flags, p->cpuid, p->pf_mask,
+						p->flags, p->cpuid, p->pfm,
 						p->uc_rev, p->uc, p->uc_size,
 						0, &samecpuid_list);
 			if (xx_uclist_add_print_errors(add_status)) {
@@ -2225,7 +2264,8 @@ static int xx_scan_handle_sig(uint32_t const id1, uint32_t const id2,
 }
 
 #ifdef USE_CPUID_DEVICE
-static int check_cpuid_devs(struct microcode_filter_entry ** const ucfp)
+static int check_cpuid_devs(__attribute__((unused)) uint32_t const sig,
+			    struct microcode_filter_entry ** const ucfp)
 {
 	uint32_t cpuid_buf[8];	/* two cpuid levels */
 	char cpuid_device[PATH_MAX];
@@ -2285,24 +2325,36 @@ static int check_cpuid_devs(struct microcode_filter_entry ** const ucfp)
 	};
 
 	if (i == 0 && ncpu == 0) {
-		print_msg(2, "cpuid kernel driver unavailable");
-		return 0;
+		print_err("cpuid kernel driver unavailable");
+		return -1;
 	} else if (rc) {
 		if (ncpu)
-			print_msg(1, "some processors were not scanned due to unexpected errors");
+			print_err("some processors were not scanned due to unexpected errors");
 		else
-			print_msg(2, "could not open cpuid devices");
+			print_err("could not open any cpuid devices");
 	}
 
 	if (ncpu)
 		print_msg(2, "checked the signature of %u processor(s)", ncpu);
 
-	return 0;
+	return rc;
 }
 #else
-static int check_cpuid_devs(__attribute__((unused)) struct microcode_filter_entry ** const uc_cpu)
+
+/* this hurts a lot less on very big systems... */
+static int check_cpuid_devs(uint32_t sig, struct microcode_filter_entry ** const ucfp)
 {
-	print_msg(2, "assuming all processors have the same signature");
+	unsigned int i;
+
+	print_msg(2, "assuming all processors have the same type, family and model");
+	sig |= 0xf;
+	for (i = 0; i < 0x10; i++) {
+		if (add_filter_to_list(sig, 0, 0, ucfp) == ENOMEM) {
+			print_err("out of memory");
+			return ENOMEM;
+		}
+		sig--;
+	}
 	return 0;
 }
 #endif
@@ -2331,24 +2383,28 @@ static int scan_system_processors(void)
 
 	switch (xx_scan_handle_sig(id1, id2, id3, sig, &uc_cpu)) {
 	case 0:
-		filter_list_allow = 0;
-		rc = check_cpuid_devs(&uc_cpu);
+		rc = check_cpuid_devs(sig, &uc_cpu);
 		break;
 	case ENXIO:
-		filter_list_allow = 0;
-		print_msg(2, "running on a non-Intel processor");
+		print_msg(1, "running on a non-Intel processor");
 		break;
 	default:
 		rc = 1;
 	}
 
+	if (!rc) {
+		filter_list_allow = 0;
+	} else {
+		print_warn("cpuid signature scan failed, ignoring incomplete results...");
+	}
+
 	if (uc_cpu) {
 		/* tie linked lists */
 		add_filter_list_to_list(&uc_filter_list, uc_cpu);
 		uc_cpu = NULL;
 	}
 
-	return rc;
+	return (rc > 0) ? rc : 0;
 }
 
 /* Command line processing */
@@ -2366,7 +2422,7 @@ static const char program_version[] =
 	"A PARTICULAR PURPOSE.";
 static const char program_bug_address[] = PACKAGE_BUGREPORT;
 
-static char cmdline_doc[] =
+static const char cmdline_doc[] =
 	PROGNAME " - Tool to manipulate Intel IA32/X86_64 microcode bundles\n"
 
 	"\v"
@@ -2407,7 +2463,7 @@ enum {
 	IUCODE_ARGP_WRITENAMEDALL,
 };
 
-static struct argp_option cmdline_options[] = {
+static const struct argp_option cmdline_options[] = {
 	{ NULL, 'h', NULL, 0, "Give this help list", -1 },
 
 	{ "quiet",   'q', NULL, 0, "Quiet operation",                1 },
@@ -2514,7 +2570,7 @@ static struct argp_option cmdline_options[] = {
 
 	{ 0 },
 };
-static char cmdline_nonarg_doc[] = "[[-t<type>] filename] ...";
+static const char cmdline_nonarg_doc[] = "[[-t<type>] filename] ...";
 
 static int new_filename(const char * const fn,
 			intel_ucode_file_type_t ftype)
